import os
import pandas as pd
import itertools
import sys
import os
BASE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../../'))
if BASE_DIR not in sys.path:
    sys.path.append(BASE_DIR)
from src.evaluation import get_metrics
from src.utils.slidingWindows import find_length_rank
import time
import torch
import numpy as np
import random
import time #to cool down processes before continuing
import warnings 
from sklearn.exceptions import UndefinedMetricWarning #ignore warnings

# seeding
seed = 2024
torch.manual_seed(seed)
torch.cuda.manual_seed(seed)
torch.cuda.manual_seed_all(seed)
np.random.seed(seed)
random.seed(seed)
torch.backends.cudnn.benchmark = False
torch.backends.cudnn.deterministic = True

print("CUDA available: ", torch.cuda.is_available())
print("cuDNN version: ", torch.backends.cudnn.version())

# 1. Hyperparameter-Optimierung
# returns best hyperparameter for trained file
''''
Args:
run_model: function that is already implemented in https://github.com/TheDatumOrg/TSB-AD/blob/main/TSB_AD/model_wrapper.py
data: preprocessed train data from one time series file
hyperparams: dict with list of hyperparameters to test
    for lOF like:   hyperparams = {
                        "n_neighbors": [10, 20, 30],
                        "metric": ["minkowski", "euclidean"],
        
                    }
model_type: unsupervised or semi-supervised

Unsupervise_AD_Pool = ['FFT', 'SR', 'IForest', 'LOF', 'POLY', 'MatrixProfile', 'PCA','KNN','KMeansAD','KShapeAD']
Semisupervise_AD_Pool = ['SAND','OCSVM','AutoEncoder', 'CNN', 'LSTMAD', 'USAD', 'OmniAnomaly']
'''

def hyperparameter_optimization(run_model, data,label,train_data, hyperparams,slidingWindow,param_combinations, model_type = 'unsupervised'):
    #ignore warnings for hyperparameter tuning
    warnings.simplefilter("ignore", category=UndefinedMetricWarning)
    #warnings.simplefilter("ignore", category=np.RankWarning)
    best_params = None
    best_score = float("-inf")

    #test each combination of hyperparameter and save best ones
    for param_set in param_combinations:
        params = dict(zip(hyperparams.keys(), param_set))

        #get scores of model
        if model_type == "unsupervised":  #for models who only gets evaluated on one dataset
            scores = run_model(data, **params)

        else:#if training and test data are nessecary
            scores = run_model(train_data, data, **params)

        #function from autors creates from scores the evaluation: returns
        # '''{
        #     'AUC-PR':float,
        #     'AUC-ROC':float,
        #     'VUS-PR':float,
        #     'VUS-ROC':float,
        #     'Standard-F1':float,
        #     'PA-F1':float,
        #    'Event-based-F1':float,
        #     'R-based-F1':float,
        #     'Affiliation-F':float, 
        #     'Precision': float, 
        #     'Recall': float
        # }'''

        metrics = get_metrics(scores, labels=label, slidingWindow=slidingWindow)
        auc_pr = metrics.get("AUC-PR", 0)# get value from auc-pr

        #check if score is better as best score
        if auc_pr > best_score:
            best_score = auc_pr
            best_params = params

    return best_params

# 2. Training- und Testfunktion
# Args:
# run_model: function that returns a scroe to evaluate
# params_grid_list: list of dict with optimated hyperparameters from set of hyperparameters and file_name: [{ params: {key:value}, file_name:str,train_data:list,data:list, label:list, sliding_window:int}]
# train_data: preprocessed trainings data
# test_data: if run method requires test data preprocessed
# model_type: declares if test data is nessecary
#Returns list of:
#"duration": [],
#     "file_name": str,
#     "params": {},
#     "group": str,
#     "point anomaly": str,
#     "seq anomaly": str,
#     "train_data":list,
#     "data":list, 
#     "label":list, 
#     "sliding_window":int
#           "metrics" : {
        #     'AUC-PR':float,
        #     'AUC-ROC':float,
        #     'VUS-PR':float,
        #     'VUS-ROC':float,
        #     'Standard-F1':float,
        #     'PA-F1':float,
        #     'Event-based-F1':float,
        #     'R-based-F1':float,
        #     'Affiliation-F':float,
        #     'Recall':float,
        #     'Precision' float,
        # }
def train_test_model(run_model, grid_list,path_to_folder, model_type="unsupervised"):
    results = grid_list.copy()

    #train a model for each file and store metrics, file_name and hyperparameters 
    for idx  in range(len(grid_list)):
        #measure duration time for each file
        start = time.time()
        time_series = grid_list[idx]
        file_name = time_series['file_name']
        #get hyperparameters for current time series
        params = dict(time_series['params'])
        data = time_series['data']
        label = time_series['label']
        if model_type == "unsupervised":  #for models who only gets evaluated on one dataset
            scores = run_model(data, **params)

        else:#if training and test data are nessecary
            train_data = time_series['train_data']
            
            scores = run_model(train_data, data, **params)
        metrics = get_metrics(scores, labels=label, slidingWindow=time_series['sliding_window'])
        results[idx]['metrics']= metrics.copy()
        end = time.time()
        duration = end-start
        results[idx]['duration'] = duration
        #save group of file
        results[idx]['group'] = file_name.split('_')[1]

        #check if file includes only point anomalies
        point_path = os.path.join(path_to_folder, 'point')
        results[idx]['point anomaly'] = os.path.isfile(os.path.join(point_path, file_name))
        #check if file includes only sequence anomalies
        seq_path = os.path.join(path_to_folder, 'seq')
        results[idx]['seq anomaly'] = os.path.isfile(os.path.join(seq_path, file_name))
    #returns metrics for evaluation
    return results

#
#Ouptput: [{file_name:str,train_data:list,data:list, label:list, sliding_window:int}]
def preprocess_data(data_path:str):
    data_list = []
    for filename in os.listdir(data_path):
        file_path = os.path.join(data_path, filename)
        df = pd.read_csv(file_path).dropna()
        data = df.iloc[:, 0:-1].values.astype(float)
        label = df['Label'].astype(int).to_numpy()

        slidingWindow = find_length_rank(data[:,0].reshape(-1, 1), rank=1)
        train_index = filename.split('.')[0].split('_')[-3]
        data_train = data[:int(train_index), :]

        data_list.append({
            'file_name':filename,
            'train_data':data_train,
            'data':data, 
            'label':label, 
            'sliding_window':slidingWindow
        })

    return data_list

#convert grid_list to results output for creating a dataframe
# OUTPUT:
# {
#     "params": [],
#     "file_name": [],
#     "duration": [],
#     "group": [],
#     "point anomaly": [],
#     "seq anomaly": [],
#     "AUC-PR": [],
#     "AUC-ROC": [],
#     "VUS-PR": [],
#     "VUS-ROC": [],
#     "Standard-F1": [],
#     "PA-F1": [],
#     "Event-based-F1": [],
#     "R-based-F1": [],
#     "Affiliation-F": []
#      "Precision": [],
#      "Recall": []
# }
    
def create_dict_from_array(input_array):
    #Initialize new dict for a dataframe
    result_dict = {
        "params": [],
        "file_name": [],
        "duration": [],
        "group": [],
        "point anomaly": [],
        "seq anomaly": [],
        "AUC-PR": [],
        "AUC-ROC": [],
        "VUS-PR": [],
        "VUS-ROC": [],
        "Standard-F1": [],
        "PA-F1": [],
        "Event-based-F1": [],
        "R-based-F1": [],
        "Affiliation-F": [],
        "Recall":[],
        "Precision": []
    }
    
    #Write single values with same index to lists in dict at same index
    for entry in input_array:
        result_dict["params"].append(entry["params"])
        result_dict["file_name"].append(entry["file_name"])
        result_dict["duration"].append(entry["duration"])
        result_dict["group"].append(entry["group"])
        result_dict["point anomaly"].append(entry["point anomaly"])
        result_dict["seq anomaly"].append(entry["seq anomaly"])

        #put metrics on same level
        metrics = entry["metrics"]
        result_dict["AUC-PR"].append(metrics["AUC-PR"])
        result_dict["AUC-ROC"].append(metrics["AUC-ROC"])
        result_dict["VUS-PR"].append(metrics["VUS-PR"])
        result_dict["VUS-ROC"].append(metrics["VUS-ROC"])
        result_dict["Standard-F1"].append(metrics["Standard-F1"])
        result_dict["PA-F1"].append(metrics["PA-F1"])
        result_dict["Event-based-F1"].append(metrics["Event-based-F1"])
        result_dict["R-based-F1"].append(metrics["R-based-F1"])
        result_dict["Affiliation-F"].append(metrics["Affiliation-F"])
        result_dict["Recall"].append(metrics["Recall"])
        result_dict["Precision"].append(metrics["Precision"])
    
    return result_dict


# 3. Main-Funktion
#Modeltype = 'unsupervised' if no test data required, else 'semi-supervised'
def main(run_model, hyperparams, model_name, data_folders = '../../../data/', model_type='unsupervised', output_dir = '../../../docs/evaluation/'):
    # train and test with all time series
    path_data_all = data_folders + 'train/all'
    path_to_folder = data_folders + 'train/'

    #go through all files in train data and store additionally informations about group and anomaly types
    grid_List = []
    # print(f"Start Processing files")
    #reutrns [{file_name:str,train_data:list,data:list, label:list, sliding_window:int}]
    file_data_dict_list = preprocess_data(path_data_all)    

    # print('Start Hyperparameter Tuning')
    # Creates combinations of all hyperparameters
    param_combinations = list(itertools.product(*hyperparams.values()))
    #find hyperparametrs for each file:
    for time_series in file_data_dict_list:
        filename = time_series['file_name']
        data = time_series['data']
        train_data = time_series['train_data']
        label = time_series['label']
        sliding_window = time_series['sliding_window']

        # Hyperparameter-Optimierung
        best_params = hyperparameter_optimization(run_model, data,label,train_data, hyperparams,sliding_window,param_combinations,model_type)
        # print(f"File: {filename}, best hyperparameter: {best_params}")
        #grid_list looks like this: [{ params: {key:value}, file_name:str,train_data:list,data:list, label:list, sliding_window:int}]
        grid_List.append({'file_name':filename,'params':best_params,'train_data':train_data,'data':data,'label':label,'sliding_window':sliding_window})

    # print('Start Training and evaluation')
    results_list = train_test_model(run_model,grid_List,path_to_folder, model_type)
    #convert list of dicts to a dict with lists
    results = create_dict_from_array(results_list)

    #check duration time
    duration = sum(results['duration'])
    # print(f'Job done. Duration: {duration:.3f} seconds')

    #store results for each model by folder 
    results_df = pd.DataFrame(results)
    output_csv = os.path.join(output_dir,model_name + '.csv')
    results_df.to_csv(output_csv, index=False)
    # print(f"Ergebnisse gespeichert in: {output_csv}")
